import itertools
import pytest
import numpy as np
import pandas as pd

from gama.utilities.metrics import Metric, all_metrics, scoring_to_metric


def _test_metric(metric, y_true, y_pred, optimal_score, prediction_score):
    def N_1_array(list_):
        return np.asarray(list_).reshape(-1, 1)

    accepted_formats = [np.asarray, N_1_array, pd.Series, pd.DataFrame]
    for truth_format, prediction_format in itertools.product(accepted_formats, accepted_formats):
        if ((y_true.ndim == 2 and y_true.shape[1] > 1 and truth_format in [pd.Series, N_1_array])
                or (y_pred.ndim == 2 and y_pred.shape[1] > 1 and prediction_format in [pd.Series, N_1_array])):
            # If a (N, K)-array is provided, we don't squash it to (N,1)
            continue

        truth = truth_format(y_true)
        prediction = prediction_format(y_pred)
        assert optimal_score == pytest.approx(metric.score(truth, truth))
        assert prediction_score == pytest.approx(metric.score(truth, prediction))


def test_accuracy_numeric():
    accuracy_metric = Metric.from_string('accuracy')
    y_true = np.asarray([1, 0, 0, 0, 1])
    y_1_mistake = np.asarray([1, 1, 0, 0, 1])
    _test_metric(accuracy_metric, y_true, y_1_mistake, optimal_score=1.0, prediction_score=0.8)


def test_accuracy_string():
    accuracy_metric = Metric.from_string('accuracy')
    y_true_str = np.asarray([str(x) for x in [1, 0, 0, 0, 1]])
    y_1_mistake_str = np.asarray([str(x) for x in [1, 1, 0, 0, 1]])
    _test_metric(accuracy_metric, y_true_str, y_1_mistake_str, optimal_score=1.0, prediction_score=0.8)


def test_logloss_numeric():
    log_loss_metric = Metric.from_string('log_loss')
    y_true = np.asarray([1, 0, 0, 0, 1])
    y_1_mistake_ohe = np.asarray([[0, 1], [0, 1], [1, 0], [1, 0], [0, 1]])
    one_mistake_logloss = 6.907755278982137
    _test_metric(log_loss_metric, y_true, y_1_mistake_ohe, optimal_score=0, prediction_score=one_mistake_logloss)

    y_true_ohe = np.asarray([[0, 1], [1, 0], [1, 0], [1, 0], [0, 1]])
    y_probabilities = np.asarray([[0.1, 0.9], [0.2, 0.8], [0.7, 0.3], [0.95, 0.05], [0.1, 0.9]])
    probabilities_logloss = 0.44562543641520713
    _test_metric(log_loss_metric, y_true_ohe, y_probabilities, optimal_score=0, prediction_score=probabilities_logloss)


def test_all_metrics_instantiate():
    for metric in all_metrics:
        Metric.from_string(metric)


def test_scoring_to_metric_mixed():
    metrics = list(all_metrics)
    mixed_metrics = [Metric.from_string(metric) for metric in metrics[:2]] + metrics[2:]
    scoring_to_metric(mixed_metrics)



import dash_core_components as dcc
import dash_html_components as html

from gama.visualization.app import dash_app
from gama.visualization.apps.dashboard_page import dashboard_page, dashboard_header

app_title = 'GAMA Dashboard'


upload_box = html.Div(
    id='upload-container',
    children=[
        dcc.Upload(
            id='upload-box',
            children=html.Div([html.A('Select or drop log(s).')]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'display': 'inline-block'
            },
            multiple=True
        )
    ],
    style=dict(width=f'{len("Select or drop log(s).")}em', display='inline-block', float='right')
)

top_bar = html.Div(
    id='top-bar',
    children=[
        html.H1(children=f"{app_title}", style={'float': 'left', 'width': f'30%', 'display': 'inline-block'}),
        html.Div(id='header-box', children=[], style=dict(float='left', width='40%')),
        upload_box,
 ])

dash_app.layout = html.Div([
    dcc.Location(id='url', refresh=False),
    top_bar,
    html.Div(id='page-content')
])

dash_app.layout['page-content'].children = dashboard_page
dash_app.layout['header-box'].children = dashboard_header


if __name__ == '__main__':
    dash_app.run_server(debug=True)



from collections import defaultdict
from typing import Dict, Any

import sklearn

from gama.genetic_programming.components import Primitive, Terminal, DATA_TERMINAL


def pset_from_config(configuration):
    """ Create a pset for the given configuration dictionary.

    Given a configuration dictionary specifying operators (e.g. sklearn
    estimators), their hyperparameters and values for each hyperparameter,
    create a gp.PrimitiveSetTyped that contains:

        - For each operator a primitive
        - For each possible hyperparameter-value combination a unique terminal

    Side effect: Imports the classes of each primitive.

    Returns the given Pset.
    """

    pset = defaultdict(list)
    parameter_checks = {}

    shared_hyperparameter_types = {}
    # We have to make sure the str-keys are evaluated first: they describe shared hyperparameters
    # We can not rely on order-preserving dictionaries as this is not in the Python 3.5 specification.
    sorted_keys = reversed(sorted(configuration.keys(), key=lambda x: str(type(x))))
    for key in sorted_keys:
        values = configuration[key]
        if isinstance(key, str):
            # Specification of shared hyperparameters
            for value in values:
                pset[key].append(Terminal(value=value, output=key, identifier=key))
        elif isinstance(key, object):
            # Specification of operator (learner, preprocessor)
            hyperparameter_types = []
            for name, param_values in sorted(values.items()):
                # We construct a new type for each hyperparameter, so we can specify
                # it as terminal type, making sure it matches with expected
                # input of the operators. Moreover it automatically makes sure that
                # crossover only happens between same hyperparameters.
                if isinstance(param_values, list) and not param_values:
                    # An empty list indicates a shared hyperparameter
                    hyperparameter_types.append(name)
                elif name == "param_check":
                    # This allows users to define illegal hyperparameter combinations, but is not a terminal.
                    parameter_checks[key.__name__] = param_values[0]
                else:
                    hyperparameter_types.append(key.__name__ + '.' + name)
                    for value in param_values:
                        pset[key.__name__ + '.' + name].append(
                            Terminal(value=value, output=name, identifier=key.__name__ + '.' + name))

            # After registering the hyperparameter types, we can register the operator itself.
            transformer_tags = ["DATA_PREPROCESSING", "FEATURE_SELECTION", "DATA_TRANSFORMATION"]
            if (issubclass(key, sklearn.base.TransformerMixin) or
                    (hasattr(key, 'metadata') and key.metadata.query()["primitive_family"] in transformer_tags)):
                pset[DATA_TERMINAL].append(Primitive(input=hyperparameter_types, output=DATA_TERMINAL, identifier=key))
            elif (issubclass(key, sklearn.base.ClassifierMixin) or
                  (hasattr(key, 'metadata') and key.metadata.query()["primitive_family"] == "CLASSIFICATION")):
                pset["prediction"].append(Primitive(input=hyperparameter_types, output="prediction", identifier=key))
            elif (issubclass(key, sklearn.base.RegressorMixin) or
                  (hasattr(key, 'metadata') and key.metadata.query()["primitive_family"] == "REGRESSION")):
                pset["prediction"].append(Primitive(input=hyperparameter_types, output="prediction", identifier=key))
            else:
                raise TypeError("Expected {} to be either subclass of "
                                "TransformerMixin, RegressorMixin or ClassifierMixin.".format(key))
        else:
            raise TypeError('Encountered unknown type as key in dictionary.'
                            'Keys in the configuration should be str or class.')

    return pset, parameter_checks


def merge_configurations(c1, c2):
    """ Takes two configurations and merges them together. """
    # Should refactor out 6 indentation levels
    merged: Dict[Any, Any] = defaultdict(lambda: None, c1)
    for algorithm, hyperparameters2 in c2.items():
        if algorithm not in merged:
            merged[algorithm] = hyperparameters2
        else:
            hyperparameters1 = merged[algorithm]
            if isinstance(hyperparameters1, list) and isinstance(hyperparameters2, list):
                #  they hyperparameters shared across algorithms
                merged[algorithm] = list(set(hyperparameters1 + hyperparameters2))
            else:
                for hyperparameter, values in hyperparameters2.items():
                    if hyperparameter not in hyperparameters1:
                        hyperparameters1[hyperparameter] = values
                    else:
                        values1 = hyperparameters1[hyperparameter]
                        if isinstance(values1, dict) and isinstance(values, dict):
                            hyperparameters1[hyperparameter] = {**values1, **values}
                        elif isinstance(values1, type(values)):
                            hyperparameters1[hyperparameter] = list(set(list(values1) + list(values)))
                        else:
                            raise TypeError(f'Could not merge values of {algorithm}.{hyperparameter}:'
                                            f'{hyperparameters1} vs. {hyperparameters2}')
    return merged



from gama.logging.GamaReport import GamaReport


def test_gamareport_from_log():
    """ Test that a GamaReport can be constructed from a log. """
    # We refer to a static log, this makes it independent of other unit tests, but it also makes it
    # independent of the actual changes in gama logging. This is bad in the sense that if logging changes
    # and the static file is not updated, we won't catch it. This is good in the sense that it achieves unit test
    # independence and backwards incompatability of changes to GamaReport are immediately caught if tested on the
    # old log first.
    # Perhaps we can/should link to the log file used in the documentation.
    log_file = 'tests/data/random_search.log'
    report = GamaReport(logfile=log_file, name=None)
    assert report.name == log_file
    assert 'RandomSearch()' == report.search_method
    assert 3 == len(report.phases)
    assert (['preprocessing', 'search', 'postprocess']
            == list(map(lambda t: t[0], report.phases)))
    assert (['default', 'RandomSearch', 'NoPostProcessing']
            == list(map(lambda t: t[1], report.phases)))
    assert report.method_data is None, "Random Search has no method data associated."


def test_gamareport_asha_from_log():
    """ Test that a GamaReport can be constructed from a log and retrieve ASHA specific information. """
    log_file = 'tests/data/asha.log'
    report = GamaReport(logfile=log_file, name=None)
    assert report.name == log_file
    assert 'AsynchronousSuccessiveHalving()' == report.search_method
    assert 3 == len(report.phases)
    assert (['preprocessing', 'search', 'postprocess']
            == list(map(lambda t: t[0], report.phases)))
    assert (['default', 'AsynchronousSuccessiveHalving', 'NoPostProcessing']
            == list(map(lambda t: t[1], report.phases)))
    assert report.method_data is not None, "ASHA has method data associated."


def test_gamareport_asyncEA_from_log():
    """ Test that a GamaReport can be constructed from a log. """
    log_file = 'tests/data/amazon_asyncEA.log'
    report = GamaReport(logfile=log_file, name=None)
    assert report.name == log_file
    assert 'AsyncEA()' == report.search_method
    assert 3 == len(report.phases)
    assert (['preprocessing', 'search', 'postprocess']
            == list(map(lambda t: t[0], report.phases)))
    assert (['default', 'AsyncEA', 'NoPostProcessing']
            == list(map(lambda t: t[1], report.phases)))
    assert report.method_data is None, "AsyncEA has no method data associated."



from gama.genetic_programming.crossover import random_crossover, crossover_primitives, _shared_terminals, crossover_terminals
from .unit_fixtures import pset, BernoulliNBStandardScaler, MultinomialNBRobustScaler, GaussianNB


def test_shared_terminals(BernoulliNBStandardScaler, MultinomialNBRobustScaler, GaussianNB):
    """ Test shared terminals are found, if they exist. """
    def cmp_len(n, generator):
        return n == len(list(generator))

    assert cmp_len(0, _shared_terminals(BernoulliNBStandardScaler, BernoulliNBStandardScaler, value_match='different'))
    assert cmp_len(2, _shared_terminals(BernoulliNBStandardScaler, BernoulliNBStandardScaler, value_match='equal'))
    assert cmp_len(2, _shared_terminals(BernoulliNBStandardScaler, BernoulliNBStandardScaler, value_match='all'))

    assert cmp_len(1, _shared_terminals(BernoulliNBStandardScaler, MultinomialNBRobustScaler, value_match='different'))
    assert cmp_len(1, _shared_terminals(BernoulliNBStandardScaler, MultinomialNBRobustScaler, value_match='equal'))
    assert cmp_len(2, _shared_terminals(BernoulliNBStandardScaler, MultinomialNBRobustScaler, value_match='all'))

    assert cmp_len(0, _shared_terminals(BernoulliNBStandardScaler, GaussianNB, value_match='all'))


def test_crossover_primitives(BernoulliNBStandardScaler, MultinomialNBRobustScaler):
    """ Two individuals of at least length 2 should produce two new individuals with crossover. """
    ind1_copy, ind2_copy = BernoulliNBStandardScaler.copy_as_new(), MultinomialNBRobustScaler.copy_as_new()
    # Cross-over is in-place
    crossover_primitives(BernoulliNBStandardScaler, MultinomialNBRobustScaler)
    # Both parents and children should be unique
    all_individuals = [BernoulliNBStandardScaler, MultinomialNBRobustScaler, ind1_copy, ind2_copy]
    assert 4 == len({ind.pipeline_str() for ind in all_individuals})
    assert ind1_copy.pipeline_str() != BernoulliNBStandardScaler.pipeline_str()


def test_crossover_terminal(BernoulliNBStandardScaler, MultinomialNBRobustScaler):
    """ Two individuals with shared Terminals should produce two new individuals with crossover. """
    ind1_copy, ind2_copy = BernoulliNBStandardScaler.copy_as_new(), MultinomialNBRobustScaler.copy_as_new()
    # Cross-over is in-place
    crossover_terminals(BernoulliNBStandardScaler, MultinomialNBRobustScaler)
    # Both parents and children should be unique
    assert len({ind.pipeline_str() for ind in [
        BernoulliNBStandardScaler, MultinomialNBRobustScaler, ind1_copy, ind2_copy]}) == 4
    assert ind1_copy.pipeline_str() != BernoulliNBStandardScaler.pipeline_str()


def test_crossover(BernoulliNBStandardScaler, MultinomialNBRobustScaler):
    """ Two eligible individuals should produce two new individuals with crossover. """
    ind1_copy, ind2_copy = BernoulliNBStandardScaler.copy_as_new(), MultinomialNBRobustScaler.copy_as_new()
    # Cross-over is in-place
    random_crossover(BernoulliNBStandardScaler, MultinomialNBRobustScaler)
    # Both parents and children should be unique
    assert len({ind.pipeline_str() for ind in [
        BernoulliNBStandardScaler, MultinomialNBRobustScaler, ind1_copy, ind2_copy]}) == 4
    assert ind1_copy.pipeline_str() != BernoulliNBStandardScaler.pipeline_str()



from abc import ABC
from typing import List, Dict, Tuple, Any, Union, Callable

import pandas as pd

from gama.genetic_programming.operator_set import OperatorSet
from gama.genetic_programming.components import Individual
from gama.utilities.generic.timekeeper import TimeKeeper


class BaseSearch(ABC):
    """ All search methods should be derived from this class.
    This class should not be directly used to configure GAMA.
    """

    def __init__(self):
        # hyperparameters can be used to safe/process search hyperparameters
        self._hyperparameters: Dict[str, Tuple[Any, Any]] = dict()
        self.output: List[Individual] = []

    def __str__(self):
        # Not sure if I should report actual used hyperparameters (i.e. include default), or only those set by user.
        user_set_hps = {parameter: set_value
                        for parameter, (set_value, default) in self._hyperparameters.items()
                        if set_value is not None}
        hp_configuration = ','.join([f"{name}={value}" for (name, value) in user_set_hps.items()])
        return f"{self.__class__.__name__}({hp_configuration})"

    @property
    def hyperparameters(self) -> Dict[str, Any]:
        """ Hyperparameter (name, value) pairs value determined by user > dynamic default > static default.

         Dynamic default values only considered if `dynamic_defaults` has been called.
         """
        return {parameter: set_value if set_value is not None else default
                for parameter, (set_value, default) in self._hyperparameters.items()}

    def _overwrite_hyperparameter_default(self, hyperparameter: str, value: Any):
        set_value, default_value = self._hyperparameters[hyperparameter]
        self._hyperparameters[hyperparameter] = (set_value, value)

    def dynamic_defaults(
            self,
            x: pd.DataFrame,
            y: Union[pd.DataFrame, pd.Series],
            time_limit: int) -> None:
        """ Set hyperparameter defaults taking into account dataset characteristics and time-constraints.

        Called before `search`.

        Parameters
        ----------
        x: pandas.DataFrame
            Features of the data.
        y: pandas.DataFrame or pandas.Series
            Labels of the data.
        time_limit: int
            Time in seconds available for search and selecting dynamic defaults.
            There is no need to adhere to this explicitly, a `stopit.utils.TimeoutException` will be raised.
            However, the time-limit might be an important factor in setting hyperparameter values.
        """
        # updates self.hyperparameters defaults
        raise NotImplementedError("Must be implemented by child class.")

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        """ Execute search as configured. Set `output` field of this class to the best Individuals.

        Parameters
        ----------
        operations: OperatorSet
            Has methods to create new individuals, evaluate individuals and more.
        start_candidates: List[Individual]
            A list of individuals to be considered before all others.
        """
        raise NotImplementedError("Must be implemented by child class.")


def _check_base_search_hyperparameters(
        toolbox,
        output: List[Individual],
        start_candidates: List[Individual]
) -> None:
    """ Checks that search hyperparameters are valid.

    :param toolbox:
    :param output:
    :param start_candidates:
    :return:
    """
    if not isinstance(start_candidates, list):
        raise TypeError(f"'start_population' must be a list but was {type(start_candidates)}")
    if not all(isinstance(x, Individual) for x in start_candidates):
        raise TypeError(f"Each element in 'start_population' must be Individual.")



from typing import NamedTuple, Tuple

Fitness = NamedTuple("Fitness",
                     [("values", Tuple),
                      ("start_time", int),
                      ("wallclock_time", int),
                      ("process_time", int)])



""" Contains full system tests for GamaRegressor """
import numpy as np
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error

from gama.postprocessing import EnsemblePostProcessing
from gama.utilities.generic.stopwatch import Stopwatch
from gama import GamaRegressor

FIT_TIME_MARGIN = 1.1
TOTAL_TIME_S = 60

# While we could derive statistics dynamically, we want to know if any changes ever happen, so we save them statically.
boston = dict(
    name='boston',
    load=load_boston,
    test_size=127,
    base_mse=81.790
)


def _test_gama_regressor(gama, X_train, X_test, y_train, y_test, data, metric):
    with Stopwatch() as sw:
        gama.fit(X_train, y_train)

    assert TOTAL_TIME_S * FIT_TIME_MARGIN >= sw.elapsed_time, 'fit must stay within 110% of allotted time.'

    predictions = gama.predict(X_test)
    assert isinstance(predictions, np.ndarray), 'predictions should be numpy arrays.'
    assert (data['test_size'],) == predictions.shape, 'predict should return (N,) shaped array.'

    # Majority classifier on this split achieves 0.6293706293706294
    mse = mean_squared_error(y_test, predictions)
    print(data['name'], metric, 'mse:', mse)
    assert data['base_mse'] >= mse, 'predictions should be at least as good as predicting mean.'
    gama.delete_cache()


def _test_dataset_problem(data, metric):
    X, y = data['load'](return_X_y=True)
    split_data = train_test_split(X, y, random_state=0)

    gama = GamaRegressor(random_state=0, max_total_time=TOTAL_TIME_S, scoring=metric, n_jobs=1, cache_dir=data['name'],
                         max_eval_time=300, post_processing_method=EnsemblePostProcessing(ensemble_size=5))
    _test_gama_regressor(gama, *split_data, data, metric)


def test_regression_mean_squared_error():
    """ GamaRegressor works on all-numeric data. """
    _test_dataset_problem(boston, 'mean_squared_error')


def test_missing_value_regression():
    """ GamaRegressor works when missing values are present. """
    data = boston
    metric = 'mean_squared_error'
    X, y = data['load'](return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)
    X_train[1:300:2, 0] = X_train[2:300:5, 1] = float("NaN")
    X_test[1:100:2, 0] = X_test[2:100:5, 1] = float("NaN")

    gama = GamaRegressor(random_state=0, max_total_time=TOTAL_TIME_S, scoring=metric, cache_dir=data['name']+'_missing')
    _test_gama_regressor(gama, X_train, X_test, y_train, y_test, data, metric)



"""
Defines the building blocks for Individuals.
Individuals represent machine learning pipelines in a back-end agnostic way.
An Individual can be converted to its back-end specific representation (e.g. a scikit-learn Pipeline
by calling its `pipeline` property as long as a function has been provided to convert the individual to it).

Individuals are built with:

 - Terminals. Definition of a specific value for a specific hyperparameter. Immutable.
 - Primitives. Definition of a specific algorithm; defined by Terminal input, output type and operation. Immutable.
 - PrimitiveNodes. An instanciated Primitive with specific Terminals. Mutable for easy operations (e.g. mutation).
 - Fitness. Stores information about the evaluation of the individual.
"""

from .primitive import Primitive
from .terminal import Terminal, DATA_TERMINAL
from .primitive_node import PrimitiveNode
from .individual import Individual
from .fitness import Fitness

__all__ = ['Individual']



from abc import ABC
from collections import defaultdict
import datetime
from functools import partial
import logging
import os
import random
import shutil
import time
from typing import Union, Tuple, Optional, Dict
import uuid
import warnings

import pandas as pd
import numpy as np
import stopit

import gama.genetic_programming.compilers.scikitlearn
from gama.logging.machine_logging import log_event, TOKENS
from gama.search_methods.base_search import BaseSearch
from gama.utilities.metrics import scoring_to_metric
from .utilities.observer import Observer

from gama.data import X_y_from_arff
from gama.search_methods.async_ea import AsyncEA
from gama.utilities.generic.timekeeper import TimeKeeper
from gama.logging.utility_functions import register_stream_log, register_file_log
from gama.utilities.preprocessing import define_preprocessing_steps, format_x_y
from gama.genetic_programming.mutation import random_valid_mutation_in_place
from gama.genetic_programming.crossover import random_crossover
from gama.genetic_programming.selection import create_from_population, eliminate_from_pareto
from gama.genetic_programming.operations import create_random_expression
from gama.configuration.parser import pset_from_config
from gama.genetic_programming.operator_set import OperatorSet
from gama.genetic_programming.compilers.scikitlearn import compile_individual
from gama.postprocessing import BestFitPostProcessing, BasePostProcessing
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from gama.utilities.metrics import Metric

log = logging.getLogger(__name__)

STR_NO_OPTIMAL_PIPELINE = """Gama did not yet establish an optimal pipeline.
                          This can be because `fit` was not yet called, or
                          did not terminate successfully."""
__version__ = '19.01.0'

for module_to_ignore in ["sklearn", "numpy"]:
    warnings.filterwarnings("ignore", module=module_to_ignore)


class Gama(ABC):
    """ Wrapper for the toolbox logic surrounding executing the AutoML pipeline. """

    def __init__(self,
                 scoring: Union[str, Metric, Tuple[Union[str, Metric], ...]] = 'filled_in_by_child_class',
                 regularize_length: bool = True,
                 config: Dict = None,
                 random_state: int = None,
                 max_total_time: int = 3600,
                 max_eval_time: Optional[int] = None,
                 n_jobs: int = -1,
                 verbosity: int = logging.WARNING,
                 keep_analysis_log: Optional[str] = 'gama.log',
                 cache_dir: Optional[str] = None,
                 search_method: BaseSearch = AsyncEA(),
                 post_processing_method: BasePostProcessing = BestFitPostProcessing()):
        """

        Parameters
        ----------
        scoring: str, Metric or Tuple
            Specifies the/all metric(s) to optimize towards. A string will be converted to Metric. A tuple must
            specify each metric with the same type (i.e. all str or all Metric). See :ref:`Metrics` for built-in
            metrics.

        regularize_length: bool
            If True, add pipeline length as an optimization metric (preferring short over long).

        config: a dictionary which specifies available components and their valid hyperparameter settings
            For more information, see :ref:`search_space_configuration`.

        random_state:  int or None (default=None)
            If an integer is passed, this will be the seed for the random number generators used in the process.
            However, with `n_jobs > 1`, there will be randomization introduced by multi-processing.
            For reproducible results, set this and use `n_jobs=1`.

        max_total_time: positive int (default=3600)
            Time in seconds that can be used for the `fit` call.

        max_eval_time: positive int, optional (default=None)
            Time in seconds that can be used to evaluate any one single individual.
            If None, set to 0.1 * max_total_time.

        n_jobs: int (default=-1)
            The amount of parallel processes that may be created to speed up `fit`.
            Accepted values are positive integers or -1.
            If -1 is specified, multiprocessing.cpu_count() processes are created.

        verbosity: int (default=logging.WARNING)
            Sets the level of log messages to be automatically output to terminal.

        keep_analysis_log: str, optional (default='gama.log')
            If non-empty str, specifies the path (and name) where the log should be stored, e.g. /output/gama.log.
            If `None`, no log is stored.

        cache_dir: str or None (default=None)
            The directory in which to keep the cache during `fit`. In this directory,
            models and their evaluation results will be stored. This facilitates a quick ensemble construction.

        search_method: BaseSearch (default=AsyncEA())
            Search method to use to find good pipelines. Should be instantiated.

        post_processing_method: BasePostProcessing (default=BestFitPostProcessing())
            Post-processing method to create a model after the search phase. Should be instantiated.

        """
        register_stream_log(verbosity)
        if keep_analysis_log is not None:
            register_file_log(keep_analysis_log)

        if keep_analysis_log is not None and not os.path.isabs(keep_analysis_log):
            keep_analysis_log = os.path.abspath(keep_analysis_log)

        arguments = ','.join(['{}={}'.format(k, v) for (k, v) in locals().items()
                              if k not in ['self', 'config', 'gamalog', 'file_handler', 'stdout_streamhandler']])
        log.info('Using GAMA version {}.'.format(__version__))
        log.info('{}({})'.format(self.__class__.__name__, arguments))
        log_event(log, TOKENS.INIT, arguments)

        if max_total_time is None or max_total_time <= 0:
            raise ValueError(f"max_total_time should be integer greater than zero but is {max_total_time}.")
        if max_eval_time is not None and max_eval_time <= 0:
            raise ValueError(f"max_eval_time should be None or integer greater than zero but is {max_eval_time}.")
        if n_jobs < -1 or n_jobs == 0:
            raise ValueError(f"n_jobs should be -1 or positive integer but is {n_jobs}.")
        elif n_jobs != -1:
            # AsyncExecutor defaults to using multiprocessing.cpu_count(), i.e. n_jobs=-1
            AsyncEvaluator.n_jobs = n_jobs

        if max_eval_time is None:
            max_eval_time = 0.1 * max_total_time
        if max_eval_time > max_total_time:
            log.warning(f"max_eval_time ({max_eval_time}) > max_total_time ({max_total_time}) is not allowed. "
                        f"max_eval_time set to {max_total_time}.")
            max_eval_time = max_total_time

        self._max_eval_time = max_eval_time
        self._time_manager = TimeKeeper(max_total_time)
        self._metrics: Tuple[Metric] = scoring_to_metric(scoring)
        self._regularize_length = regularize_length
        self._search_method: BaseSearch = search_method
        self._post_processing = post_processing_method

        default_cache_dir = f"{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}_{str(uuid.uuid4())[:4]}_GAMA"
        self._cache_dir = cache_dir if cache_dir is not None else default_cache_dir
        if not os.path.isdir(self._cache_dir):
            os.mkdir(self._cache_dir)

        if random_state is not None:
            random.seed(random_state)
            np.random.seed(random_state)

        self._X: Optional[pd.DataFrame] = None
        self._y: Optional[pd.DataFrame] = None
        self.model: object = None
        self._final_pop = None

        self._subscribers = defaultdict(list)
        self._observer = Observer(self._cache_dir)
        self.evaluation_completed(self._observer.update)

        self._pset, parameter_checks = pset_from_config(config)
        self._operator_set = OperatorSet(
            mutate=partial(random_valid_mutation_in_place, primitive_set=self._pset),
            mate=random_crossover,
            create_from_population=partial(create_from_population, cxpb=0.2, mutpb=0.8),
            create_new=partial(create_random_expression, primitive_set=self._pset),
            compile_=compile_individual,
            eliminate=eliminate_from_pareto,
            evaluate_callback=self._on_evaluation_completed
        )

    def _predict(self, x: pd.DataFrame):
        raise NotImplemented('_predict is implemented by base classes.')

    def predict(self, x: Union[pd.DataFrame, np.ndarray]):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            A dataframe or array with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            array with predictions of shape (N,) where N is the length of the first dimension of X.
        """
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x)
            for col in self._X.columns:
                x[col] = x[col].astype(self._X[col].dtype)
        return self._predict(x)

    def predict_arff(self, arff_file_path: str):
        """ Predict the target for input found in the ARFF file.

        Parameters
        ----------
        arff_file_path: str
            An ARFF file with the same columns as the one that used in fit.
            The target column is ignored (but must be present).

        Returns
        -------
        numpy.ndarray
            array with predictions for each row in the ARFF file.
        """
        if not isinstance(arff_file_path, str):
            raise TypeError(f"`arff_file_path` must be of type `str` but is of type {type(arff_file_path)}")
        X, _ = X_y_from_arff(arff_file_path)
        return self._predict(X)

    def score(self, x: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray]) -> float:
        """ Calculate the score of the model according to the `scoring` metric and input (x, y).

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            Data to predict target values for.
        y: pandas.Series or numpy.ndarray
            True values for the target.

        Returns
        -------
        float
            The score obtained on the given test data according to the `scoring` metric.
        """
        predictions = self.predict_proba(x) if self._metrics[0].requires_probabilities else self.predict(x)
        return self._metrics[0].score(y, predictions)

    def score_arff(self, arff_file_path: str) -> float:
        """ Calculate the score of the model according to the `scoring` metric and input in the ARFF file.

        Parameters
        ----------
        arff_file_path: string
            An ARFF file with which to calculate the score.

        Returns
        -------
        float
            The score obtained on the given test data according to the `scoring` metric.
        """
        X, y = X_y_from_arff(arff_file_path)
        return self.score(X, y)

    def fit_arff(self, arff_file_path: str, *args, **kwargs):
        """ Find and fit a model to predict the target column (last) from other columns.

        Parameters
        ----------
        arff_file_path: string
            Path to an ARFF file containing the training data.
            The last column is always taken to be the target.
        """
        X, y = X_y_from_arff(arff_file_path)
        self.fit(X, y, *args, **kwargs)

    def fit(self,
            x: Union[pd.DataFrame, np.ndarray],
            y: Union[pd.DataFrame, pd.Series, np.ndarray],
            warm_start: bool = False,
            keep_cache: bool = False):
        """ Find and fit a model to predict target y from X.

        Various possible machine learning pipelines will be fit to the (X,y) data.
        Using Genetic Programming, the pipelines chosen should lead to gradually
        better models. Pipelines will internally be validated using cross validation.

        After the search termination condition is met, the best found pipeline
        configuration is then used to train a final model on all provided data.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray, shape = [n_samples, n_features]
            Training data. All elements must be able to be converted to float.
        y: pandas.DataFrame, pandas.Series or numpy.ndarray, shape = [n_samples,]
            Target values. If a DataFrame is provided, it is assumed the first column contains target values.
        warm_start: bool (default=False)
            Indicates the optimization should continue using the last individuals of the
            previous `fit` call.
        keep_cache: bool (default=False)
            If True, keep the cache directory and its content after fitting is complete. Otherwise delete it.
        """

        with self._time_manager.start_activity('preprocessing', activity_meta=['default']):
            self._X, self._y = format_x_y(x, y)
            steps = define_preprocessing_steps(self._X, max_extra_features_created=None, max_categories_for_one_hot=10)
            self._operator_set._safe_compile = partial(compile_individual, preprocessing_steps=steps)

        fit_time = int((1 - self._post_processing.time_fraction) * self._time_manager.total_time_remaining)

        with self._time_manager.start_activity('search', time_limit=fit_time,
                                               activity_meta=[self._search_method.__class__.__name__]):
            self._search_phase(warm_start, timeout=fit_time)

        with self._time_manager.start_activity('postprocess',
                                               time_limit=int(self._time_manager.total_time_remaining),
                                               activity_meta=[self._post_processing.__class__.__name__]):
            best_individuals = list(reversed(sorted(self._final_pop, key=lambda ind: ind.fitness.values)))
            self._post_processing.dynamic_defaults(self)
            self.model = self._post_processing.post_process(
                self._X, self._y, self._time_manager.total_time_remaining, best_individuals
            )
        if not keep_cache:
            log.debug("Deleting cache.")
            self.delete_cache()

    def _search_phase(self, warm_start: bool = False, timeout: int = 1e6):
        """ Invoke the evolutionary algorithm, populate `final_pop` regardless of termination. """
        if warm_start and self._final_pop is not None:
            pop = self._final_pop
        else:
            if warm_start:
                log.warning('Warm-start enabled but no earlier fit. Using new generated population instead.')
            pop = [self._operator_set.individual() for _ in range(50)]

        deadline = time.time() + timeout
        evaluate_args = dict(evaluate_pipeline_length=self._regularize_length, X=self._X, y_train=self._y,
                             metrics=self._metrics, cache_dir=self._cache_dir, timeout=self._max_eval_time,
                             deadline=deadline)
        self._operator_set.evaluate = partial(gama.genetic_programming.compilers.scikitlearn.evaluate_individual,
                                              **evaluate_args)

        try:
            with stopit.ThreadingTimeout(timeout):
                self._search_method.dynamic_defaults(self._X, self._y, timeout)
                self._search_method.search(self._operator_set, start_candidates=pop)
        except KeyboardInterrupt:
            log.info('Search phase terminated because of Keyboard Interrupt.')

        self._final_pop = self._search_method.output
        log.debug([str(i) for i in self._final_pop[:100]])
        log.info(f'Search phase evaluated {len(self._observer._individuals)} individuals.')

    def _initialize_ensemble(self):
        raise NotImplementedError('_initialize_ensemble should be implemented by a child class.')

    def delete_cache(self):
        """ Removes the cache folder and all files associated to this instance.

        Returns
        -------
        None
        """
        while os.path.exists(self._cache_dir):
            try:
                log.info("Attempting to delete {}".format(self._cache_dir))
                shutil.rmtree(self._cache_dir)
            except OSError as e:
                if "The directory is not empty" not in str(e):
                    log.warning("Did not delete due to:", exc_info=True)
                # else ignore silently. This can occur if an evaluation process writes to cache.

    def _safe_outside_call(self, fn):
        """ Calls fn and log any exception it raises without reraising, except for TimeoutException. """
        try:
            fn()
        except stopit.utils.TimeoutException:
            raise
        except Exception:
            # We actually want to catch any other exception here, because the callback code can be
            # arbitrary (it can be provided by users). This excuses the catch-all Exception.
            # Note that KeyboardInterrupts are not exceptions and get elevated to the caller.
            log.warning("Exception during callback.", exc_info=True)
            pass
        if self._time_manager.current_activity.exceeded_limit:
            log.info("Time exceeded during callback, but exception was swallowed.")
            raise stopit.utils.TimeoutException

    def _on_evaluation_completed(self, ind):
        for callback in self._subscribers['evaluation_completed']:
            self._safe_outside_call(partial(callback, ind))

    def evaluation_completed(self, callback_function):
        """ Register a callback function that is called when new evaluation is completed.

        Parameters
        ----------
        callback_function:
            Function to call when a pipeline is evaluated, return values are ignored.
            Expected signature is: Individual -> Any
        """
        self._subscribers['evaluation_completed'].append(callback_function)



import logging

TIME_FORMAT = '%Y-%m-%d %H:%M:%S,%f'

# We also produce log messages below DEBUG level (machine parsable).
MACHINE_LOG_LEVEL = 5
gama_log = logging.getLogger('gama')
gama_log.setLevel(MACHINE_LOG_LEVEL)



#!/usr/bin/env python
import os

from setuptools import setup, find_packages

requirements = [
    'numpy>=1.14.0',
    'scipy>=1.0.0',
    'scikit-learn>=0.20.0',
    'stopit>=1.1.1',
    'liac-arff>=2.2.2',
    'category-encoders>=1.2.8'
]

test_requirements = [
    'pytest>=4.4.0',
    'pytest-mock',
    'pytest-xdist',
    'codecov',
    'pytest-cov'
]

visualization_requirements = [
    'dash==1.1.1',
    'dash-daq==0.1.0'
]

documentation_requirements = [
    'sphinx',
    'sphinx_rtd_theme'
] + visualization_requirements

all_ = requirements + visualization_requirements + documentation_requirements

with open(os.path.join("README.md")) as fid:
    README = fid.read()

setup(
    name='gama',
    version='19.08.0',
    description='A package for automated machine learning based on scikit-learn.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Pieter Gijsbers',
    author_email='p.gijsbers@tue.nl',
    url='https://github.com/PGijsbers/GAMA',
    project_urls={
        "Bug Tracker": "https://github.com/PGijsbers/gama/issues",
        "Documentation": "https://pgijsbers.github.io/gama/",
        "Source Code": "https://github.com/PGijsbers/gama",
    },
    packages=find_packages(exclude=['tests', "tests.*"]),
    install_requires=requirements,
    extras_require={
        'test': test_requirements,
        'vis': visualization_requirements,
        'doc': documentation_requirements,
        'all': all_
    },
    python_requires='>=3.6.0'
)



import base64
from typing import List, Dict, Optional, Any

import pandas as pd
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from gama.visualization.app import dash_app
from gama.logging.GamaReport import GamaReport
from gama.visualization.apps.plotting import individual_plot, aggregate_plot, plot_preset_graph

reports = {}
aggregate_dataframe: Optional[pd.DataFrame] = None

###########################################################
#                      HEADER BOX                         #
###########################################################
presets = [
    {'label': '#Pipeline by learner', 'value': 'number_pipeline_by_learner'},
    {'label': '#Pipeline by size', 'value': 'number_pipeline_by_size'},
    {'label': 'Best score over time', 'value': 'best_over_time'},
    {'label': 'Best score over iterations', 'value': 'best_over_n'},
    {'label': 'Size vs Metric', 'value': 'size_vs_metric'},
    {'label': 'Evaluation Times', 'value': 'evaluation_times_dist'},
    {'label': 'Evaluations by Rung', 'value': 'n_by_rung'},
    {'label': 'Time by Rung', 'value': 'time_by_rung'},
    {'label': 'Custom', 'value': 'custom'},
]

preset_container = html.Div(
    id='preset-container',
    children=[
        html.Div('Visualization Presets'),
        dcc.Dropdown(
            id='preset-dropdown',
            options=presets,
            value='best_over_n',
            style=dict(width='90%')
        )
    ],
    style=dict(width='50%', display='inline-block', float='left')
)

sep_agg_radio = dcc.RadioItems(
    id='sep-agg-radio',
    options=[
        {'label': 'separate', 'value': 'separate-line'},
        {'label': 'aggregate', 'value': 'aggregate'}
    ],
    value='separate-line',
    style={'width': '90%', 'display': 'inline-block'}
)

sep_agg_container = html.Div(
    id='sep_agg_container',
    children=[
        html.Div('Style'),
        sep_agg_radio
    ],
    style=dict(display='inline-block', width='50%', float='left')
)

dashboard_header = [preset_container, sep_agg_container]

###########################################################
#                      LEFT COLUMN                        #
###########################################################
dashboard_graph = dcc.Graph(id='dashboard-graph')

third_width = {'width': '30%', 'display': 'inline-block'}
plot_control_container = html.Div(
    id='plot-controls',
    children=[
        html.Div([
            html.Label('x-axis'),
            dcc.Dropdown(id='x-axis-metric')
            ],
            style=third_width
        ),
        html.Div([
            html.Label('y-axis'),
            dcc.Dropdown(id='y-axis-metric')
        ],
            style=third_width
        ),
        html.Div([
            html.Label('plot type'),
            dcc.Dropdown(
                id='plot-type',
                options=[
                    {'label': 'scatter', 'value': 'markers'},
                    {'label': 'line', 'value': 'lines'}
                ],
                value='lines'
            )],
            style=third_width
        )
    ],
    style=dict(width='80%', display='none'),
    hidden=True
)

graph_settings_container = html.Div(
    id='graph-settings-container',
    children=[plot_control_container]
)

graph_update_timer = dcc.Interval(
    id='update-timer',
    interval=2*1000  #ms
)

graph_update_trigger = dcc.Store(
    id='update-trigger'
)

visualization_container = html.Div(
    id='visualization-container',
    children=[dashboard_graph,
              graph_settings_container,
              graph_update_timer,
              graph_update_trigger],
    style={'float': 'left', 'width': '85%'}
)

###########################################################
#                      RIGHT COLUMN                       #
###########################################################
file_select = dcc.Checklist(id='select-log-checklist')

report_select_container = html.Div(
    id='report-select-container',
    children=[file_select],
    style={'width': '14%', 'float': 'right', 'padding-right': '1%'}
)

###########################################################
#                      Main Page                          #
###########################################################
dashboard_page = html.Div(
    id='dashboard-main',
    children=[
        visualization_container,
        report_select_container
    ]
)


###########################################################
#                      Callbacks                          #
###########################################################

@dash_app.callback(Output('update-trigger', 'data'),
                   [Input('update-timer', 'n_intervals')])
def update_reports(n_intervals: int):
    print('timer', n_intervals)
    updates = []
    for report in reports.values():
        if report.incomplete:
            updates.append(report.update())
    if any(updates):
        return None
    else:
        raise PreventUpdate


@dash_app.callback([Output('x-axis-metric', 'options'),
                    Output('y-axis-metric', 'options')],
                   [Input('select-log-checklist', 'value')])
def update_valid_axis_options(logs: List[str]):
    if logs is None or logs == []:
        return [], []
    shared_attributes = set(attribute
                            for logname, report in reports.items()
                            for attribute in report.evaluations.columns
                            if logname in logs)
    dropdown_options = [{'label': att, 'value': att} for att in shared_attributes]
    return dropdown_options, dropdown_options


@dash_app.callback(Output('dashboard-graph', 'figure'),
                   [Input('select-log-checklist', 'value'),
               Input('sep-agg-radio', 'value'),
               Input('x-axis-metric', 'value'),
               Input('y-axis-metric', 'value'),
               Input('plot-type', 'value'),
               Input('preset-dropdown', 'value'),
               Input('update-trigger', 'data')])
def update_graph(logs: List[str], aggregate: str = 'separate-line', xaxis: str = None, yaxis: str = None, mode: str = None, preset_value: str = None, trigger: Any = None):
    print(logs, aggregate, xaxis, yaxis, mode, preset_value)
    if preset_value == 'custom':
        if logs is None or logs == [] or xaxis is None or yaxis is None:
            title = 'Load and select a log on the right'
            plots = []
        else:
            title = f'{aggregate} plot of {len(logs)} logs'
            if aggregate == 'separate-line':
                plots = [individual_plot(reports[log], xaxis, yaxis, mode) for log in logs]
            if aggregate == 'aggregate':
                plots = aggregate_plot([reports[log] for log in logs], xaxis, yaxis)
        return {
            'data': plots,
            'layout': {
                'title': title,
                'xaxis': {'title': f'{xaxis}'},
                'yaxis': {'title': f'{yaxis}'},
                'hovermode': 'closest' if mode == 'markers' else 'x'
            }
        }
    elif logs is not None:
        filtered_aggregate = aggregate_dataframe[aggregate_dataframe.filename.isin(logs)]
        return plot_preset_graph([reports[log] for log in logs], filtered_aggregate, preset_value, aggregate)
    else:
        return {}


@dash_app.callback(Output('select-log-checklist', 'options'),
                   [Input('upload-box', 'contents')],
                   [State('upload-box', 'filename')])
def load_logs(list_of_contents, list_of_names):
    global aggregate_dataframe
    if list_of_contents is not None:
        for content, filename in zip(list_of_contents, list_of_names):
            content_type, content_string = content.split(',')
            decoded = base64.b64decode(content_string).decode('utf-8')
            log_lines = decoded.splitlines()
            report = GamaReport(log_lines=log_lines, name=filename)
            reports[filename] = report

            eval_copy = report.evaluations.copy()
            eval_copy['search_method'] = report.search_method
            if aggregate_dataframe is None:
                eval_copy['log_no'] = 0
                eval_copy['filename'] = filename
                aggregate_dataframe = eval_copy
            else:
                eval_copy['log_no'] = len(aggregate_dataframe['log_no'].unique())
                eval_copy['filename'] = filename
                aggregate_dataframe = pd.concat([aggregate_dataframe, eval_copy])
        return [{'label': logname, 'value': logname} for logname in reports]
    return []


@dash_app.callback(Output('plot-controls', 'style'),
                   [Input('preset-dropdown', 'value')],
                   [State('plot-controls', 'style')])
def toggle_plot_controls(preset, plot_controls_style):
    if preset == 'custom':
        plot_controls_style['display'] = 'inline-block'
    else:
        plot_controls_style['display'] = 'none'
    return plot_controls_style



import logging
from typing import List, Callable

from gama.genetic_programming.components import Individual
from gama.utilities.generic.paretofront import ParetoFront

log = logging.getLogger(__name__)


class Observer(object):
    
    def __init__(self, id_, with_log=False):
        self._with_log = with_log

        self._multiple_pareto_fronts = False
        self._overall_pareto_front = ParetoFront(get_values_fn=lambda ind: ind.fitness.values)
        self._current_pareto_front = ParetoFront(get_values_fn=lambda ind: ind.fitness.values)

        self._pareto_callbacks = []

        self._individuals = []
        self._individuals_since_last_pareto_update = 0

        self._evaluation_filename = str(id_)+'_evaluations.csv'

    def _record_individual(self, ind):
        with open(self._evaluation_filename, 'a') as fh:
            to_record = [str(ind.fitness.time),
                         str(ind.fitness.values[0]),
                         str(ind.fitness.values[1]),
                         str(ind)]
            fh.write(';'.join(to_record) + '\n')

    def update(self, ind):
        log.debug("Evaluation;{:.4f};{};{}".format(ind.fitness.wallclock_time, ind.fitness.values, ind))
        self._individuals.append(ind)
        if self._with_log:
            self._record_individual(ind)

        updated = self._current_pareto_front.update(ind)
        if updated:
            self._individuals_since_last_pareto_update = 0
            log.info("Current pareto-front updated with individual with wvalues {}.".format(ind.fitness.values))
        else:
            self._individuals_since_last_pareto_update += 1

        updated = self._overall_pareto_front.update(ind)
        if updated and self._multiple_pareto_fronts:
            self._update_pareto_front(ind)
            log.info("Overall pareto-front updated with individual with wvalues {}.".format(ind.fitness.values))

    def reset_current_pareto_front(self):
        self._current_pareto_front.clear()
        self._individuals_since_last_pareto_update = 0
        self._multiple_pareto_fronts = True

    def best_n(self, n: int) -> List[Individual]:
        """ Return the best n individuals observed based on the first optimization criterion.

        Parameters
        ----------
        n: int
            the number of individuals to return

        Returns
        -------
        List[Individual]
            A list of up to n individuals for which the score on the first criterion is the best.
            Returns less than n individuals if less than n have been evaluated.
        """
        best_pipelines = sorted(self._individuals, key=lambda x: (-x.fitness.values[0], str(x)))
        return best_pipelines[:n]

    def _update_pareto_front(self, ind):
        for callback in self._pareto_callbacks:
            callback(ind)

    def on_pareto_updated(self, fn: Callable[[Individual], None]):
        """ Register a callback function that is called when the Pareto-front is updated.

        Parameters
        ----------
        fn: Callable[[Individual], None]
            Function to call when the pareto front is updated. Expected signature is: ind -> None
        """
        self._pareto_callbacks.append(fn)



import pandas as pd

from .gama import Gama
from gama.configuration.regression import reg_config


class GamaRegressor(Gama):
    """ Wrapper for the toolbox logic executing the AutoML pipeline for regression.
    Same parameters as GamaClassifier. """

    def __init__(self, config=None, scoring='neg_mean_squared_error', *args, **kwargs):
        """ """
        # Empty docstring overwrites base __init__ doc string.
        # Prevents duplication of the __init__ doc string on the API page.

        if not config:
            config = reg_config
        super().__init__(*args, **kwargs, config=config, scoring=scoring)

    def _predict(self, x: pd.DataFrame):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe the same number of columns as that of X of `fit`.

        Returns
        -------
        numpy.ndarray
            A numpy array with predictions. The array is of shape (N,) where N is the length of the
            first dimension of X.
        """
        return self.model.predict(x)



import pytest

import pandas as pd
from sklearn.datasets import load_digits
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, log_loss

from gama import GamaClassifier


@pytest.fixture
def gamaclassifier():
    gc = GamaClassifier(random_state=0, max_total_time=60)
    yield gc
    gc.delete_cache()


def _gama_on_digits(gama):
    X, y = load_digits(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

    # Add checks on individuals (reproducibility)
    gama.fit(X_train, y_train)

    # Add checks
    y_pred = gama.predict(X_test)
    y_proba = gama.predict_proba(X_test)

    assert log_loss(y_test, y_proba) == gama.score(X_test, y_test)
    assert log_loss(y_test, y_proba) == gama.score(X_test, pd.Series(y_test))
    assert log_loss(y_test, y_proba) == gama.score(X_test, LabelEncoder().fit_transform(y_test.reshape(-1, 1)))


def test_full_system_single_core(gamaclassifier):
    from gama.utilities.generic.async_evaluator import AsyncEvaluator
    AsyncEvaluator.n_jobs = 1
    _gama_on_digits(gamaclassifier)


def test_full_system_multi_core(gamaclassifier):
    from gama.utilities.generic.async_evaluator import AsyncEvaluator
    AsyncEvaluator.n_jobs = 2
    _gama_on_digits(gamaclassifier)



import pytest
from gama import GamaClassifier
from gama.genetic_programming.components import Individual
from gama.configuration.testconfiguration import clf_config
from gama.genetic_programming.compilers.scikitlearn import compile_individual

@pytest.fixture
def pset():
    gc = GamaClassifier(config=clf_config, scoring='accuracy')
    gc.delete_cache()
    return gc._pset


@pytest.fixture
def GaussianNB(pset):
    return Individual.from_string("GaussianNB(data)",
                                  pset, compile_individual)


@pytest.fixture
def MultinomialNBRobustScaler(pset):
    return Individual.from_string("MultinomialNB(RobustScaler(data), alpha=1.0, fit_prior=True)",
                                  pset, compile_individual)


@pytest.fixture
def BernoulliNBStandardScaler(pset):
    return Individual.from_string("BernoulliNB(StandardScaler(data), alpha=0.1, fit_prior=True)",
                                  pset, compile_individual)


@pytest.fixture
def LinearSVC(pset):
    individual_str = """LinearSVC(data,
            LinearSVC.C=0.001,
            LinearSVC.dual=True,
            LinearSVC.loss='squared_hinge',
            LinearSVC.penalty='l2',
            LinearSVC.tol=1e-05)"""
    individual_str = ''.join(individual_str.split()).replace(',', ', ')
    return Individual.from_string(individual_str, pset, None)


@pytest.fixture
def RandomForestPipeline(pset):
    individual_str = """RandomForestClassifier(
            FeatureAgglomeration(
                    data,
                    FeatureAgglomeration.affinity='l2',
                    FeatureAgglomeration.linkage='complete'
                    ),   
            RandomForestClassifier.bootstrap=True,
            RandomForestClassifier.criterion='gini',
            RandomForestClassifier.max_features=0.6,
            RandomForestClassifier.min_samples_leaf=7,  
            RandomForestClassifier.min_samples_split=6, 
            RandomForestClassifier.n_estimators=100)"""
    individual_str = ''.join(individual_str.split()).replace(',', ', ')

    return Individual.from_string(individual_str, pset, None)



import itertools
import logging
import multiprocessing as mp
import queue
import sys

from gama.logging.MachineLogFileHandler import MachineLogFileHandler

gama_log = logging.getLogger('gama')


def register_stream_log(verbosity):
    previously_registered_handler = [handler for handler in gama_log.handlers if hasattr(handler, 'tag')]
    if len(previously_registered_handler) > 0:
        gama_log.debug("Removing StreamHandlers registered by previous GAMA instance(s).")
        gama_log.handlers = [handler for handler in gama_log.handlers
                             if not (hasattr(handler, 'tag') and isinstance(handler, logging.StreamHandler))]

    stdout_streamhandler = logging.StreamHandler(sys.stdout)
    stdout_streamhandler.tag = 'machine_set'
    stdout_streamhandler.setLevel(verbosity)
    gama_log.addHandler(stdout_streamhandler)


def register_file_log(filename):
    if any([isinstance(handler, MachineLogFileHandler) for handler in gama_log.handlers]):
        gama_log.debug("Removing FileHandlers registered by previous GAMA instance(s).")
        gama_log.handlers = [handler for handler in gama_log.handlers if not isinstance(handler, MachineLogFileHandler)]
    gama_log.addHandler(MachineLogFileHandler(filename))


class MultiprocessingLogger(object):
    """ Stores log messages to be written to a log later. This is helpful when communicating log messages from
    auxiliary processes to the main process.
    """

    def __init__(self):
        manager = mp.Manager()
        self._queue = manager.Queue()

    def info(self, msg):
        self._queue.put((logging.INFO, msg))

    def debug(self, msg):
        self._queue.put((logging.DEBUG, msg))

    def warning(self, msg):
        self._queue.put((logging.WARNING, msg))

    def error(self, msg):
        self._queue.put((logging.WARNING, msg))

    def log(self, level, msg):
        self._queue.put((level, msg))

    def flush_to_log(self, log):
        # According to the official documentation, Queue.Empty is not reliable, so we just poll the queue until empty.
        for i in itertools.count():
            try:
                level, message = self._queue.get(block=False)
                log.log(level, message)
            except queue.Empty:
                break

        if i > 0:
            log.debug("Flushed {} log messages.".format(i))



from typing import Tuple, List, Optional, Callable, Any


class ParetoFront(object):
    """  A Pareto front is a list of tuples in which no one tuple is dominated by another. """

    def __init__(self,
                 start_list: Optional[List[Tuple[Any, ...]]] = None,
                 get_values_fn: Optional[Callable[[Any], Tuple[Any, ...]]] = None):
        """
        Parameters
        ----------
        start_list: list, optional (default=None).
            List of items of which to calculate the Pareto front.
        get_values_fn: Callable, optional (default=None)
            Function that takes an item and returns a tuple of values such that each should be maximized.
            If left None, it is assumed that items are already tuples of which each value should be maximized.
        """
        self._get_values_fn = get_values_fn

        self._front = []
        if start_list:
            for item in start_list:
                self.update(item)

        self._iterator_index = 0

    def _get_item_value(self, item):
        if self._get_values_fn is not None:
            return self._get_values_fn(item)
        else:
            return item

    def update(self, new_item: Any):
        """ Updates the Pareto front with new_item if it dominates any current item in the Pareto front.

        Parameters
        ----------
        new_item: Any
            Item to be evaluated for submission to the Pareto front.
            Either a Tuple that matches the arity of items already in the Pareto front,
            or an object from which such a Tuple can be extracted with the `get_values_fn` provided on `__init__`.

        Returns
        -------
        bool
            True if the Pareto front is updated, False otherwise.
        """
        if not self._front:
            self._front.append(new_item)
            return True

        values = self._get_item_value(new_item)
        old_arity = len(self._get_item_value(self._front[0]))
        if old_arity != len(values):
            raise ValueError("Arity of added tuple must match that of the ones in the Pareto front. Front tuples have "
                             "arity {} and new tuple has arity {}.".format(len(self._front[0]), len(values)))

        to_remove = []

        # Check for each point whether new_item dominates it, it gets dominated, or neither.
        for old_item in self._front:
            old_values = self._get_item_value(old_item)
            if all(v1 <= v2 for v1, v2 in zip(values, old_values)):
                # old_item dominates this new_item
                return False
            elif all(v1 >= v2 for v1, v2 in zip(values, old_values)):
                # new_item dominates this old_item
                to_remove.append(old_item)
            # else: Neither dominates nor gets dominated by old_item, which means new_item is still a candidate

        # new_item was not dominated by any Pareto-front item, which means it gets added.
        # so old items that new_item dominates must be removed.
        for item in to_remove:
            self._front.remove(item)
        self._front.append(new_item)
        return True

    def clear(self):
        """ Removes all items from the Pareto front."""
        self._front = []

    def __iter__(self):
        self._iterator_index = 0
        return self

    def __next__(self):
        if self._iterator_index >= len(self._front):
            raise StopIteration

        self._iterator_index += 1
        return self._front[self._iterator_index - 1]

    def __len__(self):
        return len(self._front)

    def __getitem__(self, item):
        return self._front[item]

    def __str__(self):
        return str(self._front)

    def __repr__(self):
        if self._get_values_fn is None:
            return 'ParetoFront({})'.format(str(self._front))
        else:
            return 'ParetoFront({}, get_values_fn="{}")'.format(str(self._front), self._get_values_fn.__name__)



from gama.postprocessing.base_post_processing import BasePostProcessing
from gama.postprocessing.best_fit import BestFitPostProcessing
from gama.postprocessing.ensemble import EnsemblePostProcessing


class NoPostProcessing(BasePostProcessing):
    """ Does nothing, no time will be reserved for post-processing. """
    def __init__(self, time_fraction: float = 0.0):
        super().__init__(time_fraction)

    def post_process(self, *args, **kwargs):
        return None


__all__ = ['NoPostProcessing', 'BestFitPostProcessing', 'EnsemblePostProcessing']



import time
import pytest
from gama.utilities.generic.stopwatch import Stopwatch


ROUND_ERROR = 0.02


def test_stopwatch_initialization_zero():
    """ Test that elapsed time is 0 if stopwatch is not started yet. """
    sw = Stopwatch()
    assert pytest.approx(0, abs=ROUND_ERROR) == sw.elapsed_time


def test_stopwatch_elapsed_time_while_running():
    """ Tests that elapsed_time increments as expected while running. """
    with Stopwatch() as sw:
        assert pytest.approx(0, abs=ROUND_ERROR) == sw.elapsed_time
        time.sleep(1)
        assert pytest.approx(1, abs=ROUND_ERROR) == sw.elapsed_time


def test_stopwatch_elapsed_time_after_running():
    """ Tests that time elapsed is stored after exiting the context. """
    with Stopwatch() as sw:
        time.sleep(1)
    time.sleep(1)
    assert pytest.approx(1, abs=ROUND_ERROR) == sw.elapsed_time



from typing import Callable, NamedTuple, Tuple

# Defines an operator which takes input and produces output, e.g. a preprocessing or classification algorithm.
Primitive = NamedTuple("Primitive",
                       [("input", Tuple[str]),
                        ("output", str),
                        ("identifier", Callable)])


def primitive__str__(primitive) -> str:
    """ str: e.g. "FastICA" """
    return primitive.identifier.__name__


Primitive.__str__ = primitive__str__
Primitive.__repr__ = primitive__str__
